//================================================================================================
// $Id: symtab.h 6309 2023-01-16 20:19:42Z aron $
//
// Project:
//		Welltype Parser
// Contents:
//		Symbol table
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#ifndef WTPARSER_SYMTAB_H
#define WTPARSER_SYMTAB_H

#include "parser_pre.h"
#include <string>
#include <map>
#include <vector>

namespace wtast
{
class Type;
} // namespace wtast

namespace wtparser
{
class ParserHelper;

class SymTab
{
	struct Variable
	{
		Variable() { }

		Variable(const std::string & n, wtast::Type* t, UniqueId u, bool tg, wt_u32 d, wtast::AstNode::Variable::Class vc)
			: name(n)
			, type(t)
			, unique_id(u)
			, tagged(tg)
			, depth(d)
			, varclass(vc)
		{
			// ok
		}

		std::string                     name;
		wtast::Type*                    type;
		UniqueId                        unique_id;
		bool                            tagged;
		wt_u32                          depth;
		wtast::AstNode::Variable::Class varclass;
	};

	typedef std::map<std::string, Variable> varmap_t;

	struct Frame
	{
		Frame(const std::string & name) : name(name)
		{
			// ok
		}

		Frame(const std::string & name, const Frame & frame)
			: name(name)
			, vars(frame.vars)
		{
			// ok
		}

		Variable* Find(const std::string & name)
		{
			varmap_t::iterator it = vars.find(name);

			if(it!=vars.end())
			{
				return &it->second;
			}

			return 0;
		}

		void Insert(const std::string & name, wtast::Type* type, UniqueId unique_id, bool tagged, wt_u32 depth, wt_u32 loadsize, wtast::AstNode::Variable::Class varclass);
		bool HideByUniqueId(UniqueId unique_id);

		std::string name;
		varmap_t    vars;
	};

	typedef std::vector<Frame> framestack_t;

	ParserHelper* mHelper;
	Frame         mGlobal;
	framestack_t  mFrameStack;

	Frame & GetInsertionFrame(wtast::AstNode::Variable::Class varclass)
	{
		if(wtast::AstNode::Variable::GLOBAL==varclass)
		{
			return mGlobal;
		}

		return mFrameStack.back();
	}

public:

	struct Resolver
	{
		wtast::Type*                    type;
		std::string                     name;
		std::string                     frame_name;
		unsigned int                    unique_id;
		bool                            tagged;
		wtast::AstNode::Variable::Class varclass;
	};

public:
	SymTab(ParserHelper* helper);

	~SymTab()
	{
		// ok
	}

	static const std::string UnnamedFrame;

	bool EnterBlock();
	bool EnterBlock(const std::string &);
	void LeaveBlock();
	bool AddVariable(wtast::Type*, const std::string &, UniqueId, bool tagged, wtast::AstNode::Variable::Class varclass);
	bool Resolve(Resolver & resolver, const std::string & var_name, const std::string & frame_name);
	void HideVariable(UniqueId unique_id);
	wt_u32 CurrentDepth() const { return mFrameStack.size(); }
};

} // wtparser

#endif // WTPARSER_SYMTAB_H
