#ifndef WTRUNTIME_CTX_H
#define WTRUNTIME_CTX_H

#include "config.h"
#include "types.h"
#include "frame.h"
#include "dynstack.h"

#define WTR_CTX_STACK_PUSH(_ctx_, _stack_, _type_, _data_) \
	do \
	{ \
		const _type_ tmp____ = _data_; \
		(_ctx_)->_stack_ -= sizeof(_type_); \
		wtr_vma_write(vm, (_ctx_)->_stack_, &tmp____, sizeof(_type_)); \
	} \
	while(0)

#define WTR_CTX_STACK_POP(_ctx_, _stack_, _size_, _data_) \
	do \
	{ \
		wtr_vma_read(vm, (_ctx_)->_stack_, &(_data_), (_size_)); \
		(_ctx_)->_stack_ += (_size_); \
	} \
	while(0)

#define WTR_CTX_STACK_PEEK(_ctx_, _stack_, _size_, _data_) \
	do \
	{ \
		wtr_vma_read(vm, (_ctx_)->_stack_, &(_data_), (_size_)); \
	} \
	while(0)

#define WTR_CTX_DS_SIZE(_ctx_)  ((wt_uintptr)((_ctx_)->ds_hi - (_ctx_)->dsp))
#define WTR_CTX_DS_SPACE(_ctx_) ((wt_uintptr)((_ctx_)->dsp - (_ctx_)->ds_lo))
#define WTR_CTX_CS_SIZE(_ctx_)  ((wt_uintptr)((_ctx_)->cs_hi - (_ctx_)->csp))
#define WTR_CTX_CS_SPACE(_ctx_) ((wt_uintptr)((_ctx_)->csp - (_ctx_)->cs_lo))

#define WTR_CTX_DS_PUSH(_ctx_,_type_,_data_) WTR_CTX_STACK_PUSH(_ctx_, dsp, _type_, _data_)
#define WTR_CTX_CS_PUSH(_ctx_,_type_,_data_) WTR_CTX_STACK_PUSH(_ctx_, csp, _type_, _data_)
#define WTR_CTX_DS_PEEK(_ctx_,_type_,_data_) WTR_CTX_STACK_PEEK(_ctx_, dsp, sizeof(_type_), _data_)
#define WTR_CTX_DS_POP(_ctx_,_type_,_data_)  WTR_CTX_STACK_POP(_ctx_, dsp, sizeof(_type_), _data_)
#define WTR_CTX_CS_POP(_ctx_,_type_,_data_)  WTR_CTX_STACK_POP(_ctx_, csp, sizeof(_type_), _data_)

struct wtr_vm;
struct wtr_program;
struct wtr_sym_func;

typedef enum wtr_runresult (*WTR_PFN_IOCALLNEXT)(struct wtr_context* const ctx);
typedef void (*WTR_PFN_RETURNACTION)(struct wtr_context* const ctx);

struct wtr_context
{
	struct wtr_vm*         vm;             /* the virtual machine this context belongs to */
	enum wtr_runresult     run;            /* actual state */
	wt_exception           last_exception; /* last unhandled exception */
	wt_vmptr               dsp;            /* Data Stack Pointer (grows toward lower address) */
	wt_vmptr               csp;            /* Call Stack Pointer (grows toward lower address) */
	struct wtr_callframe   frame;          /* actual call frame */
	struct wtr_dynstack    tos;            /* Temporary Object Stack (TOS) */
	struct wtr_dynstack    rtos;           /* Reverse TOS (content of the CSOS) */
	wt_vmptr               ds_lo;          /* the data stack starts here (this is the lowest address; inclusive) */
	wt_vmptr               ds_hi;          /* the data stack ends here (this is the highest address; exclusive) */
	wt_vmptr               cs_lo;          /* the call stack starts here (this is the lowest address; inclusive) */
	wt_vmptr               cs_hi;          /* the call stack ends here (this is the highest address; exclusive) */
	wt_vmptr               stack_base;     /* the allocated buffer for the stack (VM-space) or the preallocated memory */
	wt_u32                 stack_size;     /* size of the actual stack memory block */
	void*                  user_pointer;   /* opaque user pointer */
	enum wtr_iocall        last_iocall;    /* last pending I/O call */
	wt_u16                 iocall_argc;    /* argument count for I/O call */
	wt_u16                 iocall_retc;    /* number of return values expected from the I/O call */
	const wt_u32*          iocall_retv;    /* return value type array for I/O call */
	void*                  iocall_npriv;   /* private pointer for 'iocall_next' */
	WTR_PFN_IOCALLNEXT     iocall_next;    /* follow-up function for successful I/O call */
	wt_u8                  iocall_wlen;    /* 'iocall_wbuf' length (used in WTIOC_FWRITE follow-up) */
	wt_u8                  iocall_wbuf[6]; /* buffer for additional write in WTIOC_FWRITE follow-up */
	wt_u8                  stack_fixed;    /* non-zero if the stack was preallocated */
	char*                  sbuf;           /* temporary string buffer */
	wt_u32                 sbuf_size;      /* size of 'sbuf' */
	wt_u32                 value_arr_size; /* number of elements in 'value_array' */
	wt_value*              value_array;    /* the value array for external calls */
	struct wtr_sym_func*   called_func;    /* function that was called */
	struct wtr_program*    called_prog;    /* function was called in this program */
	wt_u32                 retc;           /* number of returned values (main{}, ctor{} and dtor{} has no return value) */
	struct wtr_extcallinfo extcallinfo;    /* information for the external function call */
	wt_value               iocall_argv[WTR_IOCALL_ARGMAX]; /* arguments for I/O call */
	WTR_PFN_RETURNACTION   return_action;  /* function to be called when the context enters WTRR_RETURNED */
};

wt_error wtr_ctx_invoke_ctor(struct wtr_context* const ctx,
	struct wtr_program* const prog);

wt_error wtr_ctx_invoke_dtor(struct wtr_context* const ctx,
	struct wtr_program* const prog);

wt_bool wtr_ctx_resize_value_array(struct wtr_context* const ctx,
	wt_uint32_t size);

/* Pop values from the DataStack determined by 'types' into 'values' (and */
/* fill type info in 'values'). Return zero (aka UnknownException) on */
/* success. */
wt_exception wtr_ctx_pop_values(struct wtr_context* const ctx,
	wt_u32 nvalues, wt_value* const values, const wt_u32* types);

/* Return zero (aka UnknownException) on success. */
wt_exception wtr_ctx_push_values(struct wtr_context* const ctx,
	const wt_u32 nvalues, wt_value* const values);

#endif /* WTRUNTIME_CTX_H */
