#ifndef WTRUNTIME_DEBUG_H
#define WTRUNTIME_DEBUG_H

#include "config.h"

#if WTRUNTIME_DEBUG

#  include "wtruntime.h"
#  include <assert.h>
#  include <stddef.h>
#  include <string.h>

#if 0
/* display executed instructions on standard output */
#  define DISPLAY_EXECUTED_INSTRUCTIONS 1
#endif
#if 0
/* display data stack content if DISPLAY_EXECUTED_INSTRUCTIONS enabled */
#  define DISPLAY_DATASTACK_AFTER_INSTRUCTIONS 1
#endif
#if 0
/* enables logging of object allocation and deallocation */
#  define LOG_OBJECT_ALLOCATIONS 1
#endif
#if 0
/* display object allocation/deallocation on standard output */
#  define DISPLAY_ALLOCATIONS 1
#endif
#if 0
/* add magic number after the object allocation */
#  define ADD_OBJECT_MAGIC 1
#endif
#if 0
/* add a magic number inside the object structures */
#  define ADD_INLINE_OBJECT_MAGIC 1
#endif

#endif /* WTRUNTIME_DEBUG */

#ifdef ADD_OBJECT_MAGIC
#  define OBJECT_MAGIC_LENGTH_FOR_STR 16
#  define OBJECT_MAGIC_LENGTH_FOR_SEQ 16
#  define OBJECT_MAGIC_LENGTH_FOR_REC 16
#else /* ADD_OBJECT_MAGIC */
#  define OBJECT_MAGIC_LENGTH_FOR_STR 0
#  define OBJECT_MAGIC_LENGTH_FOR_SEQ 0
#  define OBJECT_MAGIC_LENGTH_FOR_REC 0
#endif /* ADD_OBJECT_MAGIC */

#ifdef LOG_OBJECT_ALLOCATIONS
#  include <stdio.h>
struct wtr_ptrset;
extern struct wtr_ptrset* allocated;

struct wtr_ptrset* wtr_ptrset_create(size_t size);
void wtr_ptrset_delete(struct wtr_ptrset* const ptrset);
int wtr_ptrset_has(struct wtr_ptrset* const ptrset, const void* const key); /* returns non-zero if element exists */
int wtr_ptrset_add(struct wtr_ptrset* const ptrset, const void* const key); /* key shall not be freed up */
void wtr_ptrset_remove(struct wtr_ptrset* const ptrset, const void* const key);

#  ifdef DISPLAY_ALLOCATIONS
#    define DISPLAY_ALLOCATION(fmt, ...) printf(fmt, ## __VA_ARGS__)
#  else /* DISPLAY_ALLOCATIONS */
#    define DISPLAY_ALLOCATION(...)
#  endif /* DISPLAY_ALLOCATIONS */

#  define ALLOCATE_BTREE \
		do \
		{ \
			if(!allocated) \
			{ \
				allocated = wtr_ptrset_create(256*sizeof(void*)); \
			} \
		} \
		while(0)

#  define LOG_ALLOC(_name_,_ptr_,_size_) \
		do \
		{ \
			void* const ptr = _ptr_; \
			\
			ALLOCATE_BTREE; \
			\
			if(0!=wtr_ptrset_add(allocated, ptr)) \
			{ \
				printf("Reallocated %s: %p (%zu octets)\n", _name_, ptr, (_size_)); \
			} \
			\
			DISPLAY_ALLOCATION("NEW %s %p (%zu octets)\n", _name_, ptr, (_size_)); \
		} \
		while(0)

#  define LOG_FREE(_name_,_ptr_) \
		do \
		{ \
			void* const ptr = _ptr_; \
			\
			ALLOCATE_BTREE; \
			\
			DISPLAY_ALLOCATION("DEL %s %p\n", _name_, ptr); \
			\
			if(!wtr_ptrset_has(allocated, ptr)) \
			{ \
				printf("Free unallocated %s: %p\n", _name_, ptr); \
			} \
			\
			wtr_ptrset_remove(allocated, ptr); \
		} \
		while(0)

#  define ASSERT_OBJECT_EXISTENCE(_ptr_) \
		do \
		{ \
			void* const ptr = _ptr_; \
			\
			ALLOCATE_BTREE; \
			\
			assert(wtr_ptrset_has(allocated, ptr)); \
		} \
		while(0)

#else /* LOG_OBJECT_ALLOCATIONS */
#  define LOG_ALLOC(...)
#  define LOG_FREE(...)
#  define DISPLAY_ALLOCATION(...)
#  define ASSERT_OBJECT_EXISTENCE(...)
#endif /* LOG_OBJECT_ALLOCATIONS */

#ifdef ADD_OBJECT_MAGIC
void wtr_debug_check_all_bytes(void* data, char ch, size_t size, wt_bool clear);

#  define SET_OBJECT_MAGIC_STR(str) \
		memset(((char*)str)+sizeof(wtr_str)+str->length, '\xf6', OBJECT_MAGIC_LENGTH_FOR_STR)
#  define SET_OBJECT_MAGIC_SEQ(str) \
		memset(((char*)seq)+sizeof(wtr_seq)+seq->length*seq->elemsize, '\xf7', OBJECT_MAGIC_LENGTH_FOR_SEQ)
#  define SET_OBJECT_MAGIC_REC(str) \
		memset(((char*)rec)+sizeof(wtr_rec)+rec->size, '\xf8', OBJECT_MAGIC_LENGTH_FOR_REC)

#  define CHECK_OBJECT_MAGIC_STR(str) \
		wtr_debug_check_all_bytes(((char*)str)+sizeof(wtr_str)+str->length, '\xf6', OBJECT_MAGIC_LENGTH_FOR_STR, WT_TRUE)
#  define CHECK_OBJECT_MAGIC_SEQ(seq) \
		wtr_debug_check_all_bytes(((char*)seq)+sizeof(wtr_seq)+seq->length*seq->elemsize, '\xf7', OBJECT_MAGIC_LENGTH_FOR_SEQ, WT_TRUE)
#  define CHECK_OBJECT_MAGIC_REC(rec) \
		wtr_debug_check_all_bytes(((char*)rec)+sizeof(wtr_rec)+rec->size, '\xf8', OBJECT_MAGIC_LENGTH_FOR_REC, WT_TRUE)

#else /* ADD_OBJECT_MAGIC */
#  define SET_OBJECT_MAGIC_STR(...)
#  define SET_OBJECT_MAGIC_SEQ(...)
#  define SET_OBJECT_MAGIC_REC(...)
#  define CHECK_OBJECT_MAGIC_STR(...)
#  define CHECK_OBJECT_MAGIC_SEQ(...)
#  define CHECK_OBJECT_MAGIC_REC(...)
#endif /* ADD_OBJECT_MAGIC */

#ifdef ADD_INLINE_OBJECT_MAGIC
#  include "objects.h"
#  define CHECK_OBJECT_MAGIC12_STR(_obj_vma_) assert(wtr_str_is_valid(vm, (_obj_vma_)))
#  define CHECK_OBJECT_MAGIC12_SEQ(_obj_vma_) assert(wtr_seq_is_valid(vm, (_obj_vma_)))
#  define CHECK_OBJECT_MAGIC12_REC(_obj_vma_) assert(wtr_rec_is_valid(vm, (_obj_vma_)))
#else /* ADD_INLINE_OBJECT_MAGIC */
#  define CHECK_OBJECT_MAGIC12_STR(_obj_)
#  define CHECK_OBJECT_MAGIC12_SEQ(_obj_vma_)
#  define CHECK_OBJECT_MAGIC12_REC(_obj_vma_)
#endif /* ADD_INLINE_OBJECT_MAGIC */

#endif /* WTRUNTIME_DEBUG_H */
