#include "config_private.h"
#include "sym.h"
#include "symtab.h"
#include "program.h"
#include "vm.h"
#include <string.h>
#if WTRUNTIME_DEBUG
#  include <assert.h>
#endif /* WTRUNTIME_DEBUG */

struct wtr_sym* wtr_sym_new(struct wtr_vm* const vm, enum wtr_symtype symtype,
	const enum wtr_symbind symbind, struct wtr_symtab* const symtab,
	size_t size, const char* name, size_t namelen)
{
	struct wtr_sym* const sym =
		(struct wtr_sym*)wtr_vm_sys_malloc(vm, size+namelen+1);

	char* const a_name = ((char*)sym) + size;

	if(NULL!=sym)
	{
		sym->type   = symtype;
		sym->bind   = symbind;
		sym->symtab = symtab;
		sym->name   = a_name;

		memcpy(a_name, name, namelen);
		a_name[namelen] = 0;
	}

	return sym;
}

void wtr_sym_release(struct wtr_vm* const vm, struct wtr_sym* const sym)
{
	wtr_vm_sys_free(vm, sym);
}

static int cmp__uint(const wt_uint32_t lhs, const wt_uint32_t rhs)
{
	if(lhs<rhs) { return -1; }
	if(lhs>rhs) { return +1; }
	return 0;
}

static int cmp__symtype(enum wtr_symtype lhs, enum wtr_symtype rhs)
{
	if(wtr_sym_is_func(lhs) && wtr_sym_is_func(rhs))
	{
		return 0;
	}

	return cmp__uint(lhs, rhs);
}

static int cmp__type(struct wtr_vm* const vm, wt_typeid lhs, wt_typeid rhs)
{
	return cmp__uint(lhs, rhs);
}

static int cmp__typevec(struct wtr_vm* const vm, const wt_typeid* lhs,
	const wt_typeid* rhs, wt_uint32_t cnt)
{
	const wt_typeid* const end = lhs + cnt;

	while(lhs<end)
	{
		const int res = cmp__type(vm, *lhs, *rhs);

		if(0!=res)
		{
			return res;
		}

		++rhs, ++lhs;
	}

	return 0;
}

static int cmp__funcsig(struct wtr_vm* const vm,
	const struct wtr_s_funcsig* const lhs,
	const struct wtr_s_funcsig* const rhs)
{
	int res;

	const wt_uint32_t argc = lhs->argc;
	const wt_uint32_t retc = lhs->retc;

	if(0!=(res=cmp__uint(argc, rhs->argc))) { return res; }
	if(0!=(res=cmp__uint(retc, rhs->retc))) { return res; }
	if(0!=(res=cmp__uint(lhs->is_pure, rhs->is_pure))) { return res; }
	if(0!=(res=cmp__typevec(vm, lhs->argv, rhs->argv, argc))) { return res; }
	if(0!=(res=cmp__typevec(vm, lhs->retv, rhs->retv, retc))) { return res; }

	return 0;
}

static int wtr_sym_cmp__func(struct wtr_vm* const vm,
	const struct wtr_sym_func* const lhs,
	const struct wtr_sym_func* const rhs)
{
	return cmp__funcsig(vm, &lhs->sig, &rhs->sig);
}

static int wtr_sym_cmp__functype(struct wtr_vm* const vm,
	const struct wtr_sym_functype* const lhs,
	const struct wtr_sym_functype* const rhs)
{
	return cmp__funcsig(vm, &lhs->sig, &rhs->sig);
}

static int  wtr_sym_cmp__oper(
	const struct wtr_sym_oper* const lhs,
	const struct wtr_sym_oper* const rhs)
{
	int res;

	if(0!=(res=cmp__uint(lhs->precedence, rhs->precedence)))
	{
		return res;
	}

	if(0!=(res=cmp__uint(lhs->associativity, rhs->associativity)))
	{
		return res;
	}

	return 0;
}

static int wtr_sym_cmp__rec(struct wtr_vm* const vm,
	const struct wtr_sym_rec* const lhs,
	const struct wtr_sym_rec* const rhs)
{
	int res;

	const wt_uint32_t fldc = lhs->fldc;

	if(0!=(res=cmp__uint(fldc, rhs->fldc))) { return res; }

	{
		const struct wtr_s_recfield* l = lhs->fldv;
		const struct wtr_s_recfield* r = rhs->fldv;

		wt_uint32_t i;

		for(i=0;i<fldc;++i)
		{
			if(0!=(res=strcmp(l->name, r->name))) { return res; }
			if(0!=(res=cmp__type(vm, l->type, r->type))) { return res; }

			++l, ++r;
		}
	}

	return 0;
}

static int wtr_sym_cmp__enum(
	const struct wtr_sym_enum* const lhs,
	const struct wtr_sym_enum* const rhs)
{
	int res;

	const wt_uint32_t cnt = lhs->enumeratorc;

	if(0!=(res=cmp__uint(cnt, rhs->enumeratorc))) { return res; }

	{
		const char** l = lhs->enumeratorv;
		const char** r = rhs->enumeratorv;

		wt_uint32_t i;

		for(i=0;i<cnt;++i)
		{
			if(0!=(res=strcmp(*l, *r))) { return res; }

			++l, ++r;
		}
	}

	return 0;
}

static int cmp__datacons(struct wtr_vm* const vm,
	const struct wtr_s_datacons* const lhs,
	const struct wtr_s_datacons* const rhs)
{
	int res;

	const wt_uint32_t typc = lhs->typc;

	if(0!=(res=cmp__uint(typc, rhs->typc))) { return res; }
	if(0!=(res=cmp__typevec(vm, lhs->typv, rhs->typv, typc))) { return res; }

	return 0;
}

static int wtr_sym_cmp__datatype(struct wtr_vm* const vm,
	const struct wtr_sym_datatype* const lhs,
	const struct wtr_sym_datatype* const rhs)
{
	int res;

	const wt_uint32_t numcons = lhs->numcons;

	if(0!=(res=cmp__uint(numcons, rhs->numcons))) { return res; }
	if(0!=(res=cmp__uint(lhs->defcons, rhs->defcons))) { return res; }

	{
		struct wtr_s_datacons* l = lhs->conses;
		struct wtr_s_datacons* r = rhs->conses;

		wt_uint32_t i;

		for(i=0;i<numcons;++i)
		{
			if(0!=(res=cmp__datacons(vm, l, r))) { return res; }

			++l, ++r;
		}
	}

	return 0;
}

int wtr_sym_cmp(struct wtr_vm* const vm, struct wtr_sym* const lhs,
	struct wtr_sym* const rhs)
{
	const enum wtr_symtype type = lhs->type;

	/* shallow compare: type */
	{
		const int res = cmp__symtype(type, rhs->type);

		if(0!=res)
		{
			return res;
		}
	}

	/* shallow compare: name */
	{
		const int res = strcmp(lhs->name, rhs->name);

		if(0!=res)
		{
			return res;
		}
	}

	/* deep compare */

	switch(type)
	{
	case WTR_SYMTYPE_WTFUNC:
	case WTR_SYMTYPE_CFUNC:
		return wtr_sym_cmp__func(vm,
			(const struct wtr_sym_func*)lhs,
			(const struct wtr_sym_func*)rhs);

	case WTR_SYMTYPE_EXC:
		return 0;

	case WTR_SYMTYPE_OPER:
		return wtr_sym_cmp__oper(
			(const struct wtr_sym_oper*)lhs,
			(const struct wtr_sym_oper*)rhs);

/*	case WTR_SYMTYPE_TYPEAL: TODO */
/*	case WTR_SYMTYPE_SUBTYPE: TODO */

	case WTR_SYMTYPE_DATATYPE:
		return wtr_sym_cmp__datatype(vm,
			(const struct wtr_sym_datatype*)lhs,
			(const struct wtr_sym_datatype*)rhs);

	case WTR_SYMTYPE_REC:
		return wtr_sym_cmp__rec(vm,
			(const struct wtr_sym_rec*)lhs,
			(const struct wtr_sym_rec*)rhs);

	case WTR_SYMTYPE_PRIVREC:
		return 0;

	case WTR_SYMTYPE_LIMREC:
		return 0;

	case WTR_SYMTYPE_FUNCTYPE:
		return wtr_sym_cmp__functype(vm,
			(const struct wtr_sym_functype*)lhs,
			(const struct wtr_sym_functype*)rhs);

	case WTR_SYMTYPE_ENUM:
		return wtr_sym_cmp__enum(
			(const struct wtr_sym_enum*)lhs,
			(const struct wtr_sym_enum*)rhs);

	default:
#if WTRUNTIME_DEBUG
		assert(! "unhandled symtype in wtr_sym_cmp()");
#endif /* WTRUNTIME_DEBUG */
		return -1;
	}
}

static int test_type_name(void* const ctx, struct wtr_sym* const sym)
{
	const enum wtr_symtype symtype = sym->type;

	if(!wtr_sym_is_type(symtype))
	{
		return 1;
	}

	return strcmp((const char*)ctx, sym->name);
}

static int test_oper_name(void* const ctx, struct wtr_sym* const sym)
{
	const enum wtr_symtype symtype = sym->type;

	if(!wtr_sym_is_oper(symtype))
	{
		return 1;
	}

	return strcmp((const char*)ctx, sym->name);
}

static void report_unres_exc(struct wtr_vm* const vm,
	struct wtr_sym_exc* const exc)
{
	const struct wtr_loaderops* const loaderops = vm->loaderops;

	if(NULL!=loaderops)
	{
		void (*const unres_exc)(void* ctx, const char* name) =
			loaderops->unres_exc;

		if(NULL!=unres_exc)
		{
			unres_exc(vm->loaderops_ctx, exc->sym.name);
		}
	}
}

static void report_unres_type(struct wtr_vm* const vm,
	struct wtr_sym_type* const type)
{
	const struct wtr_loaderops* const loaderops = vm->loaderops;

	if(NULL!=loaderops)
	{
		void (*const unres_type)(void* ctx, const char* name) =
			loaderops->unres_type;

		if(NULL!=unres_type)
		{
			unres_type(vm->loaderops_ctx, type->sym.name);
		}
	}
}

static void report_unres_oper(struct wtr_vm* const vm,
	struct wtr_sym_oper* const oper)
{
	const struct wtr_loaderops* const loaderops = vm->loaderops;

	if(NULL!=loaderops)
	{
		void (*const unres_op)(void* ctx, const char* name) =
			loaderops->unres_op;

		if(NULL!=unres_op)
		{
			unres_op(vm->loaderops_ctx, oper->sym.name);
		}
	}
}

static void report_unres_func(struct wtr_vm* const vm,
	struct wtr_sym_func* const func)
{
	const struct wtr_loaderops* const loaderops = vm->loaderops;

	if(NULL!=loaderops)
	{
		void (*const unres_func)(void* ctx, const char* name,
			wt_uint32_t argc, wt_uint32_t retc, const wt_typeid* argv,
			const wt_typeid* retv) = loaderops->unres_func;

		if(NULL!=unres_func)
		{
			unres_func(vm->loaderops_ctx, func->sym.name, func->sig.argc,
				func->sig.retc, func->sig.argv, func->sig.retv);
		}
	}
}

static wt_error wtr_sym_resolve_type(struct wtr_vm* const vm,
	wt_typeid* const p_typeinfo, struct wtr_symtab* const symtab,
	struct wtr_sym* (*get_usertype)(struct wtr_symtab* const symtab,
	wt_uint32_t usertype_id))
{
	wt_typeid flags = WTRT_TYPEFLAGS(*p_typeinfo);

	if(WTRTF_USER & flags)
	{
		struct wtr_sym_type* const type = (struct wtr_sym_type*)get_usertype(
			symtab, WTRT_TYPEID(*p_typeinfo));

		wt_typeid typeinfo;

		if(NULL==type || !wtr_sym_is_type(type->sym.type))
		{
			/* type not found, or not a type */
			return WTE_TYPEINF;
		}

		typeinfo = type->typeinfo;

		if(0!=((WTRTF_REF|WTRTF_MUTABLE|WTRTF_SEQ) & WTRT_TYPEFLAGS(typeinfo)))
		{
			/* unsupported type flags */
			return WTE_TYPEINF;
		}

		flags &= WTRTF_REF | WTRTF_MUTABLE | WTRTF_SEQ;
		flags |= WTRTF_USER & WTRT_TYPEFLAGS(typeinfo);

		*p_typeinfo = WTRT_TYPEINFO(flags, WTRT_TYPEID(typeinfo));
	}

	return WTE_OK;
}

static wt_error wtr_sym_deep_resolve_funcsig(struct wtr_vm* const vm,
	struct wtr_s_funcsig* const sig, struct wtr_symtab* const symtab,
	struct wtr_sym* (*get_usertype)(struct wtr_symtab* const symtab,
	wt_uint32_t usertype_id))
{
	wt_error err;

	{
		wt_typeid*       typ = sig->argv;
		wt_typeid* const end = typ + sig->argc;

		while(typ<end)
		{
			if(WTE_OK!=(err=wtr_sym_resolve_type(vm, typ++, symtab,
				get_usertype)))
			{
				return err;
			}
		}
	}

	{
		wt_typeid*       typ = sig->retv;
		wt_typeid* const end = typ + sig->retc;

		while(typ<end)
		{
			if(WTE_OK!=(err=wtr_sym_resolve_type(vm, typ++, symtab,
				get_usertype)))
			{
				return err;
			}
		}
	}

	return WTE_OK;
}

static wt_error wtr_sym_deep_resolve_sym(struct wtr_vm* const vm,
	struct wtr_sym* const sym, struct wtr_symtab* const symtab,
	struct wtr_sym* (*get_usertype)(struct wtr_symtab* const symtab,
	wt_uint32_t usertype_id))
{
	wt_error err;

	switch(sym->type)
	{
	case WTR_SYMTYPE_WTFUNC:
	case WTR_SYMTYPE_CFUNC:
		{
			struct wtr_sym_func* const func = (struct wtr_sym_func*)sym;

			if(WTE_OK!=(err=wtr_sym_deep_resolve_funcsig(vm, &func->sig,
				symtab, get_usertype)))
			{
				return err;
			}
		}
		break;

	case WTR_SYMTYPE_FUNCTYPE:
		{
			struct wtr_sym_functype* const functype =
				(struct wtr_sym_functype*)sym;

			if(WTE_OK!=(err=wtr_sym_deep_resolve_funcsig(vm, &functype->sig,
				symtab, get_usertype)))
			{
				return err;
			}
		}
		break;

	case WTR_SYMTYPE_REC:
		{
			struct wtr_sym_rec* const rec = (struct wtr_sym_rec*)sym;

			struct wtr_s_recfield*       fld = rec->fldv;
			struct wtr_s_recfield* const end = fld + rec->fldc;

			for(;fld<end;++fld)
			{
				if(WTE_OK!=(err=wtr_sym_resolve_type(vm, &fld->type, symtab,
					get_usertype)))
				{
					return err;
				}
			}
		}
		break;

	case WTR_SYMTYPE_DATATYPE:
		{
			struct wtr_sym_datatype* const datatype =
				(struct wtr_sym_datatype*)sym;

			struct wtr_s_datacons*       cons  = datatype->conses;
			struct wtr_s_datacons* const c_end = cons + datatype->numcons;

			for(;cons<c_end;++cons)
			{
				wt_typeid*       typ = cons->typv;
				wt_typeid* const end = typ + cons->typc;

				while(typ<end)
				{
					if(WTE_OK!=(err=wtr_sym_resolve_type(vm, typ++, symtab,
						get_usertype)))
					{
						return err;
					}
				}
			}
		}
		break;

	default:
		break;
	}

	return WTE_OK;
}

wt_error wtr_sym_bind(struct wtr_vm* const vm, struct wtr_sym* const sym)
{
	const enum wtr_symtype symtype = sym->type;

	wt_error    err;
	wt_globalid gid;

	if(wtr_sym_is_exc(symtype))
	{
		struct wtr_sym_exc* const exc = (struct wtr_sym_exc*)sym;

		if(NULL==wtr_vm_find_sym(vm, sym))
		{
			if(WTE_OK!=(err=wtr_gidtree_add(&vm->gidtree, vm, &gid, sym)))
			{
				return err;
			}

			exc->exc_id = gid;
		}
		else
		{
			exc->exc_id = WTR_SYM_UNDEF;
		}
	}
	else
	if(wtr_sym_is_type(symtype))
	{
		struct wtr_sym_type* const type = (struct wtr_sym_type*)sym;

		if(WTR_SYMBIND_EXPORT==sym->bind)
		{
			if(WTE_OK!=(err=wtr_gidtree_add(&vm->gidtree, vm, &gid, sym)))
			{
				return err;
			}
		}
		else
		{
			if(0!=wtr_gidtree_find(&vm->gidtree, &gid, (void*)sym->name,
				test_type_name))
			{
				report_unres_type(vm, type);
				return WTE_UNRES;
			}
		}

		type->typeinfo = WTRT_TYPEINFO(WTRTF_USER, gid);
	}
	else
	if(wtr_sym_is_func(symtype))
	{
		struct wtr_sym_func* const func = (struct wtr_sym_func*)sym;

		if(WTR_SYMBIND_EXPORT==sym->bind)
		{
			if(WTE_OK!=(err=wtr_gidtree_add(&vm->gidtree, vm, &gid, sym)))
			{
				return err;
			}

			func->func_id = gid;
		}
	}
	else
	if(wtr_sym_is_oper(symtype))
	{
		struct wtr_sym_oper* const oper = (struct wtr_sym_oper*)sym;

		if(WTR_SYMBIND_EXPORT==sym->bind)
		{
			if(0==wtr_gidtree_find(&vm->gidtree, &gid, (void*)sym->name,
				test_oper_name))
			{
				return WTE_AMBIG;
			}

			if(WTE_OK!=(err=wtr_gidtree_add(&vm->gidtree, vm, &gid, sym)))
			{
				return err;
			}
		}
		else
		{
			if(0!=wtr_gidtree_find(&vm->gidtree, &gid, (void*)sym->name,
				test_oper_name))
			{
				return WTE_AMBIG;
			}
		}

		oper->oper_id = gid;
	}
#if WTRUNTIME_DEBUG
	else
	{
		assert(! "unhandled symtype in wtr_sym_bind()");
	}
#endif /* WTRUNTIME_DEBUG */

	return WTE_OK;
}

static wt_error add_ref_symtab(struct wtr_vm* const vm,
	struct wtr_symtab* const symtab, struct wtr_sym* const sym)
{
	struct wtr_symtab* const other_symtab = sym->symtab;

	if(NULL!=other_symtab)
	{
		struct wtr_vector* const ref_symtabs =
			&symtab->prog->referenced_symtabs;

		if(!wtr_vector_find(ref_symtabs, other_symtab, NULL))
		{
			if(0!=wtr_vector_push_back(ref_symtabs, vm, other_symtab))
			{
				return WTE_NOMEM;
			}
		}
	}

	return WTE_OK;
}

wt_error wtr_sym_resolve(struct wtr_vm* const vm, struct wtr_sym* const sym,
	struct wtr_symtab* const symtab,
	struct wtr_sym* (*get_usertype)(struct wtr_symtab* const symtab,
	wt_uint32_t usertype_id))
{
	const enum wtr_symtype symtype = sym->type;

	wt_error err;

	if(WTE_OK!=(err=wtr_sym_deep_resolve_sym(vm, sym, symtab, get_usertype)))
	{
		return err;
	}

	if(wtr_sym_is_exc(symtype))
	{
		struct wtr_sym_exc* const exc = (struct wtr_sym_exc*)sym;

		if(WTR_SYM_UNDEF==exc->exc_id)
		{
			struct wtr_sym_exc* const other =
				(struct wtr_sym_exc*)wtr_vm_find_sym(vm, sym);

			if(NULL==other)
			{
				report_unres_exc(vm, exc);
				return WTE_UNRES;
			}

			if(WTE_OK!=(err=add_ref_symtab(vm, symtab,
				(struct wtr_sym*)other)))
			{
				return err;
			}

			exc->exc_id = other->exc_id;
		}
	}
	else
	if(wtr_sym_is_type(symtype))
	{
		struct wtr_sym_type* const type = (struct wtr_sym_type*)sym;

		if(WTR_SYMBIND_IMPORT==sym->bind)
		{
			struct wtr_sym* const other = wtr_vm_find_sym(vm, sym);

			if(NULL==other)
			{
				report_unres_type(vm, type);
				return WTE_UNRES;
			}

			if(WTE_OK!=(err=add_ref_symtab(vm, symtab, other)))
			{
				return err;
			}
		}
	}
	else
	if(wtr_sym_is_func(symtype))
	{
		struct wtr_sym_func* const func = (struct wtr_sym_func*)sym;

		if(WTR_SYMBIND_IMPORT==sym->bind)
		{
			struct wtr_sym_func* const other =
				(struct wtr_sym_func*)wtr_vm_find_sym(vm, sym);

			if(NULL==other)
			{
				report_unres_func(vm, func);
				return WTE_UNRES;
			}

			if(WTE_OK!=(err=add_ref_symtab(vm, symtab,
				(struct wtr_sym*)other)))
			{
				return err;
			}

			func->func_id = other->func_id;
		}
	}
	else
	if(wtr_sym_is_oper(symtype))
	{
		struct wtr_sym_oper* const oper = (struct wtr_sym_oper*)sym;

		if(WTR_SYMBIND_IMPORT==sym->bind)
		{
			struct wtr_sym_oper* const other =
				(struct wtr_sym_oper*)wtr_vm_find_sym(vm, sym);

			if(NULL==other)
			{
				report_unres_oper(vm, oper);
				return WTE_UNRES;
			}

			if(WTE_OK!=(err=add_ref_symtab(vm, symtab,
				(struct wtr_sym*)other)))
			{
				return err;
			}

			oper->oper_id = other->oper_id;
		}
	}
#if WTRUNTIME_DEBUG
	else
	{
		assert(! "unhandled symtype in wtr_sym_resolve()");
	}
#endif /* WTRUNTIME_DEBUG */

	return WTE_OK;
}

void wtr_sym_unreg_gid(struct wtr_vm* const vm, struct wtr_sym* const sym)
{
	const enum wtr_symtype symtype = sym->type;

	wt_globalid gid;

	if(wtr_sym_is_exc(symtype))
	{
		struct wtr_sym_exc* const exc = (struct wtr_sym_exc*)sym;

		gid = exc->exc_id;
		exc->exc_id = WTR_SYM_UNDEF;
	}
	else
	if(wtr_sym_is_type(symtype))
	{
		struct wtr_sym_type* const type = (struct wtr_sym_type*)sym;

		const wt_typeid typeinfo = type->typeinfo;

		if(!(WTRTF_USER & WTRT_TYPEFLAGS(typeinfo)))
		{
			return;
		}

		gid = WTRT_TYPEID(typeinfo);
		type->typeinfo = WTR_SYM_UNDEF;
	}
	else
	if(wtr_sym_is_func(symtype))
	{
		struct wtr_sym_func* const func = (struct wtr_sym_func*)sym;

		gid = func->func_id;
		func->func_id = WTR_SYM_UNDEF;
	}
	else
	if(wtr_sym_is_oper(symtype))
	{
		struct wtr_sym_oper* const oper = (struct wtr_sym_oper*)sym;

		gid = oper->oper_id;
		oper->oper_id = WTR_SYM_UNDEF;
	}
	else
	{
		return;
	}

	if(sym==wtr_gidtree_get(&vm->gidtree, gid))
	{
		wtr_gidtree_clear(&vm->gidtree, gid);
	}
}
