#include "config_private.h"
#include "symtab3.h"
#include "symtab.h"
#include "sym.h"
#include "vector.h"
#include "vm.h"
#include <string.h>

struct wtr_symtab3
{
	struct wtr_symtab symtab;
	struct wtr_vector syms;
	wt_uint32_t       typecnt;
};

static struct wtr_sym* wtr_symtab3_lookup_impl(
	struct wtr_symtab3* const symtab, struct wtr_vm* const vm,
	struct wtr_sym* const sym, size_t* const out_index)
{
	const size_t num_syms = wtr_vector_size(&symtab->syms);

	if(0<num_syms)
	{
		struct wtr_sym** const start =
			(struct wtr_sym**)wtr_vector_at(&symtab->syms, 0);

		struct wtr_sym** const end = start + num_syms;

		struct wtr_sym** syms = start;

		while(syms<end)
		{
			struct wtr_sym* const my_sym = *syms++;

			if(NULL!=my_sym && 0==wtr_sym_cmp(vm, my_sym, sym))
			{
				*out_index = (syms-1) - start;
				return my_sym;
			}
		}
	}

	return NULL;
}

static struct wtr_sym* wtr_symtab3_lookup(struct wtr_symtab* const symtab,
	struct wtr_vm* const vm, struct wtr_sym* const sym)
{
	size_t index;

	return wtr_symtab3_lookup_impl((struct wtr_symtab3*)symtab, vm, sym,
		&index);
}

static void wtr_symtab3_destroy(struct wtr_symtab* const symtab0,
	struct wtr_vm* const vm)
{
	struct wtr_symtab3* const symtab = (struct wtr_symtab3*)symtab0;

	wtr_symtab_symvec_destroy(vm, &symtab->syms);
	wtr_vm_sys_free(vm, symtab0);
}

struct wtr_symtab3* wtr_symtab3_create(struct wtr_vm* const vm,
	struct wtr_program* const prog, wt_uint32_t typecnt)
{
	struct wtr_symtab3* const symtab =
		(struct wtr_symtab3*)wtr_vm_sys_malloc(vm,
		sizeof(struct wtr_symtab3));

	if(NULL==symtab)
	{
		return NULL;
	}

	memset(symtab, 0, sizeof(*symtab));

	symtab->symtab.prog    = prog;
	symtab->symtab.lookup  = wtr_symtab3_lookup;
	symtab->symtab.destroy = wtr_symtab3_destroy;
	symtab->typecnt        = typecnt;

	wtr_list_init(&symtab->symtab.link_symtabs);

	if(0!=wtr_vector_create(&symtab->syms, vm, typecnt))
	{
		goto error;
	}

	/* TODO: maybe if(0<typecnt) { wtr_vector_resize(); } */
	while(wtr_vector_size(&symtab->syms)<typecnt)
	{
		if(0!=wtr_vector_push_back(&symtab->syms, vm, NULL))
		{
			goto error;
		}
	}

	return symtab;

error:
	wtr_symtab3_destroy((struct wtr_symtab*)symtab, vm);
	return NULL;
}

wt_error wtr_symtab3_add_type(struct wtr_symtab3* const symtab,
	struct wtr_vm* const vm, struct wtr_sym* const sym,
	wt_uint32_t type_index)
{
	struct wtr_sym** const slot =
		(struct wtr_sym**)wtr_vector_at(&symtab->syms, type_index);

	if(NULL==slot)
	{
		return WTE_NOMEM;
	}

	if(NULL==(*slot))
	{
		*slot = sym;
		return WTE_OK;
	}

	if(0==wtr_sym_cmp(vm, *slot, sym))
	{
		if(WTR_SYMBIND_EXPORT==(*slot)->bind)
		{
			/* type symbol in the slot is an export, keep it */
			wtr_sym_release(vm, sym);
		}
		else
		{
			/* type symbol in the slot is an import, replace it */
			wtr_sym_release(vm, *slot);
			*slot = sym;
		}

		return WTE_OK;
	}

	return WTE_BINERR;
}

wt_error wtr_symtab3_add_symbol(struct wtr_symtab3* const symtab,
	struct wtr_vm* const vm, struct wtr_sym* const sym,
	size_t* const out_index)
{
	size_t index;
	struct wtr_sym* const old_sym = wtr_symtab3_lookup_impl(symtab, vm, sym,
		&index);

	if(NULL!=old_sym)
	{
		/* replace the new sym with the old one, exports are added first */
		wtr_sym_release(vm, sym);
		*out_index = index;
		return WTE_OK;
	}

	*out_index = wtr_vector_size(&symtab->syms);

	if(0!=wtr_vector_push_back(&symtab->syms, vm, sym))
	{
		return WTE_NOMEM;
	}

	return WTE_OK;
}

struct wtr_sym* wtr_symtab3_get_symbol(struct wtr_symtab3* const symtab,
	wt_uint32_t sym_index)
{
	struct wtr_sym** const slot =
		(struct wtr_sym**)wtr_vector_at(&symtab->syms, sym_index);

	if(NULL!=slot)
	{
		return *slot;
	}

	return NULL;
}

static struct wtr_sym* wtr_symtab3_get_usertype(
	struct wtr_symtab* const symtab, wt_uint32_t usertype_id)
{
	return wtr_symtab3_get_symbol((struct wtr_symtab3*)symtab, usertype_id);
}

wt_error wtr_symtab3_resolve_undef(struct wtr_symtab3* const symtab,
	struct wtr_vm* const vm)
{
	wt_error err = WTE_OK;

	const size_t num_syms = wtr_vector_size(&symtab->syms);

	if(0<num_syms)
	{
		struct wtr_sym** const start =
			(struct wtr_sym**)wtr_vector_at(&symtab->syms, 0);

		struct wtr_sym** const end = start + num_syms;

		struct wtr_sym** syms;

		/* bind symbols */

		for(syms=start;syms<end;++syms)
		{
			struct wtr_sym* const sym = *syms;

			if(NULL!=sym)
			{
				const wt_error bind_res = wtr_sym_bind(vm, sym);

				if(WTE_OK!=bind_res)
				{
					err = bind_res;
				}
			}
		}

		if(WTE_OK!=err)
		{
			return err;
		}

		/* resolve symbols */

		for(syms=start;syms<end;++syms)
		{
			struct wtr_sym* const sym = *syms;

			if(NULL!=sym)
			{
				const wt_error rr = wtr_sym_resolve(vm, sym,
					(struct wtr_symtab*)symtab, wtr_symtab3_get_usertype);

				if(WTE_OK!=rr)
				{
					err = rr;
				}
			}
		}
	}

	return err;
}
